/*  *********************************************************************** */
/*  *********************************************************************** */

/*  FILE main-dp.c - *SAT 1.2 */
/*  Main program */

/*  *********************************************************************** */
/*  *********************************************************************** */
/* This is main! */
#define IN_MAIN

/* Standard modules */
#include <stdio.h>
#include <sys/time.h>
#include <sys/resource.h>

/* *SAT modules */
#include "fnode.h"
#include "dag.h"
#include "dp_sat.h"
#include "satotrie.h" 
#include "monitor.h"
#include "table.h"
#include "set.h"

/* Choose the logic consistency test package */
#ifdef TRANSLATE_TO_K
#include "kdp.h"
#else

#ifdef KSAT
#include "kdp.h"
#endif

#ifdef ESAT
#include "edp.h"
#endif

#endif

/* Main variables */
extern fnode_t  *formula_as_tree;
dag_t           *dag_ptr;
image_t         *main_image_ptr;
monitor_t       *main_monitor_ptr;
int             res;
Cons_func_ptr_t Test_star_consistency;

/* Prototypes */
extern int      yyparse();
Cons_func_ptr_t Choose_test(); 
void            Translate_to_K(fnode_t *formula_fnode_ptr);
void            Set_preprocessing_flags(int *bary, int *negate);
void            Setup_memoizing(dag_t *dag_ptr);
int             Skip_params(int argc, char **argv); 
void            Help_and_exit(char **argv);
void            Display_results(monitor_t *main_monitor_ptr);
void            Preparse_cmd_line(int argc, char **argv);
void            Purge_comments();

int main(int argc, char **argv)
{
  int filec;
  int bary, negate;

  /* Initializing the monitor */
  main_monitor_ptr = Init_monitor();

  /* If no argument is supplied give help and exit starSAT */
  if (argc < 2) Help_and_exit(argv);

  /* Getting the index for the file name in command line */
  filec = Skip_params(argc, argv);

  /* Parsing the input formula------------------------------------------------ */
  if (freopen(argv[filec], "r", stdin) != NULL) { 
#ifdef PURGE_COMMENTS
    Purge_comments();
#endif
    Start_clock(main_monitor_ptr, PARSE);
    yyparse();
    Stop_clock(main_monitor_ptr, PARSE);
    /* Recover stdin!!! */
    freopen("/dev/tty", "r", stdin); 
  } else { 
    printf("File %s does not exists\n",argv[1]); 
    exit(-1); 
  } 

  /* Preprocessing------------------------------------------------------------ */
  Start_clock(main_monitor_ptr, PREPROCESS);
  Set_preprocessing_flags(&bary, &negate);
  Preparse_cmd_line(argc, argv);
  if (bary) {
    /* If the input syntax defines binary trees,
       they are converted to n-ary trees */
    formula_as_tree = Convert_binary2nary(formula_as_tree);
    Flatten_and_or(formula_as_tree);  
  }
  if (NNF_CONV) 
    /* Decide whether to push down negation or not */
    if (negate) Convert_NNF(formula_as_tree, 1); else Convert_NNF(formula_as_tree, 0);
  else {
    /* Just get rid of diamonds from the input syntax */
    Convert_dia2box(formula_as_tree);
  }
  Stop_clock(main_monitor_ptr, PREPROCESS);

#ifdef TRANSLATE_TO_K
  /* Applying a translational approach for a classical modal logic */
  Start_clock(main_monitor_ptr, TRANSLATE);
  Translate_to_K(formula_as_tree);
  Stop_clock(main_monitor_ptr, TRANSLATE);
#endif

  /* Building ---------------------------------------------------------------- */
  Start_clock(main_monitor_ptr, BUILD);
  /* Working either as a validity checker (the input formula is negated) or 
     as a satisfiability checker. */
  if (NNF_CONV || !negate)
    /* Formula was (eventually) negated by NNF conversion */ 
    dag_ptr = Make_dag(formula_as_tree, 0, LANGUAGE, CNF_CONV); 
  else
    /* Formula must be negated here */
    dag_ptr = Make_dag(Make_formula_nary(not_code, empty_code, formula_as_tree), 
		       0, LANGUAGE, CNF_CONV);  
  /* Replacing boxes with modal atoms */
  Replace_boxes(dag_ptr);  
  /* Initializing the decision algorithm */
  main_image_ptr = Init_cnf_sat(dag_ptr, GREEDY_CNF_CONV);  
  Stop_clock(main_monitor_ptr, BUILD);

  /* Eventually, command line parameters change default initialization */
  parse_cmd_line(main_image_ptr, argc, argv);

/*    Print_dag_lisp(dag_ptr); exit(0); */

#ifdef DEBUG
  Print_dag_info(dag_ptr); 
  exit(0);
#endif

  /* Starting the decision algorithm ----------------------------------------- */
  Setup_memoizing(dag_ptr);
  Image_monitor_ptr(main_image_ptr) = main_monitor_ptr;
  Test_star_consistency = Choose_test();

  Start_clock(main_monitor_ptr, DSAT); 
  res = Test_dp_sat(main_image_ptr, dag_ptr, Test_star_consistency);
  Stop_clock(main_monitor_ptr, DSAT); 

  /* Get memory allocation data and displaying results */
  Inc_param_by(main_monitor_ptr, MAX_SATO_MEMORY, Memory_count(main_image_ptr));
  Display_results(main_monitor_ptr);
  
  /* Clear things away */
  Clear_image(main_image_ptr);
  Clear_monitor(main_monitor_ptr);

  return (0);
}

Cons_func_ptr_t Choose_test()
{
#ifdef TRANSLATE_TO_K
  return (Test_K_consistency);
#else

#ifdef KSAT
  return (Test_K_consistency);
#endif

#ifdef ESAT
  return (Test_E_consistency);
#endif

#endif
}

void Setup_memoizing(dag_t *dag_ptr)
{
  Force_no_memoT();
  Force_no_memoS();


  if (CACHE_MEMO == 1) {
#ifdef OPTIMIZE
    Init_memoT(Lma(dag_ptr) + 1);
#else
    Init_memoT();
#endif
  } else if (CACHE_MEMO >= 2) {
    Init_memoS(Lma(dag_ptr));
    
    if ((CACHE_MEMO == 3) || (CACHE_MEMO == 5) || (CACHE_MEMO == 7))
      Setf_memoS(TRASH_SUBSETS); 
    if (CACHE_MEMO >= 4)
      Setf_memoS(GET_DEPENDENCIES); 
    if (CACHE_MEMO >= 6) 
      Setf_memoS(LEVEL_CACHE); 

  }

  return;
}

void Set_preprocessing_flags(int *bary, int *negate)
{
#ifdef BARY
  *bary = 1;
#else
  *bary = 0;
#endif
  
#ifdef NEGATE
  *negate = 1;
#else
  *negate = 0;
#endif
  
  return;
}


void Translate_to_K(fnode_t *formula_fnode_ptr)
{
#ifdef ESAT
  Translate_from_E_to_K(formula_fnode_ptr);
#endif
}

int Skip_params(int argc, char **argv)
{
  int i;
  
  for(i = 1; i < argc; i++) 
    if (argv[i][0] != '-') break;

  return i;
}

void Help_and_exit(char **argv)
{
  printf("Welcome to *SAT, a platform for experimenting with SAT-based procedures!\n");
  printf("Usage: %s [<Options>] <Input_file>\n\n", argv[0]);
  printf("Options are:\n");
  printf("********** SPLITTING **********\n");
  printf("   -s<n>    : choose a different splitting heuristic (n = 1..6):\n");
  printf("               n = 1 MOM strategy to all active variables\n");
  printf("               n = 2 MOM strategy to all active variables in the\n"); 
  printf("                     shortest clauses\n");
  printf("               n = 3 choose the min variable in the first shortest\n");
  printf("                      clause (sign = CHOICE1)\n");
  printf("               n = 4 choose a variable in a shortest clause using\n");
  printf("                     occurrence as the second criterion (sign = CHOICE1)\n");
  printf("               n = 5 choose a variable with the highest occurrence\n");
  printf("                     (sign = CHOICE1)\n");
  printf("               n = 6 choose a variable with the highest occurence\n");
  printf("                     (sign = ?)\n");
  printf("   -l<n>    : choose a different MOM strategy (n = 1..3 - 11..13 - 21):\n");
  printf("               n < 11       original SATO weighting:\n");
  printf("                   n=1 sign =(neg_count>pos_count)? CHOICE1 : CHOICE2;\n");
  printf("                   n=2 sign = FF\n");
  printf("                   n=3 sign = TT\n");
  printf("               11 <= n < 21 clauses lenght does not matter (sign as n - 10)\n");
  printf("               n = 21       Boehm's weight and sign\n");
  printf("   -v       : splitting bias becomes CHOICE1 = TT, CHOICE2 = FF\n\n");

  printf("********** OPTIMIZATIONS (PROPOSITIONAL LOGIC ONLY!!)  **********\n");
  printf("   -p<n>    : enable pure literal detection     \n");
  printf("              n=1 on added variables only       \n");
  printf("              n=2 on any variable               \n");
  printf("   -g<n>    : enable propositional backjumping  \n");        
  printf("   -h       : enable relaxation to horn clauses \n\n");        

  printf("********** INTERNAL CNF CONVERSION **********\n");
  printf("   -k       : choose the conversion type\n");
  printf("              n=0 standard CNF conversion\n");
  printf("              n=1 optimized CNF conversion (also triggers -n, -p1)\n");
  printf("   -n       : pushing down negations\n");
  printf("   -d       : don't prefer splitting on independent variables\n\n");

  printf("********** MODAL LOGICS **********\n");
  printf("   -e       : switch modal early pruning\n");
  printf("               n=0 early pruning is off\n");
  printf("               n=1 early pruning is on\n");
  printf("   -b<n>    : enable modal backjumping\n");
  printf("               n=1 try to find smallest inconsistent kernel\n");
  printf("               n=2 discard irrelevant atoms only\n");
  printf("   -m<n>    : enable memoizing:\n");
  printf("               n=2 set memoizing\n");
  printf("               n=3 as 2 but also try to be conservative, i.e.,\n");
  printf("                   try to trash useless data first\n");    
  printf("               n=4 as 2 but also store and test dependencies\n");
  printf("               n=5 as 3 but also store and test dependencies\n");
  printf("               n=6 as 4 but also differentiate the cache into modal levels\n");
  printf("               n=7 as 5 but also differentiate the cache into modal levels\n");
  printf("   -w<n>    : cache memory size (a multiple of 32 is good!)\n");

  printf("********** MISCELLANEA **********\n");
  printf("   -r       : give a rich output\n");
  printf("   -o<sec>  : enable a timeout (disable any timeout if 0)\n\n");

  printf("Defaults are: -s2 -l1 (CHOICE1 = FF, CHOICE2 = TT)\n");
  printf("              -g0, pure literal and relaxation to horn cl. are disabled\n");
  printf("              splitting on dependent variables is disabled\n");
  printf("              -k1, negations are pushed down\n");
  printf("              -e1, backjumping is disabled, -m6, -w512\n");
  printf("              -o1000, output is compact\n\n");
  
  exit (0);
}

void Display_results(monitor_t *main_monitor_ptr)
{
  if (VERBOSE) {
    printf("The formula was %s\n", (res ? "satisfiable" : "unsatisfiable"));
    printf("--------------------TIMINGS\n");
    printf("Parse time        %s\n", Print_clock(main_monitor_ptr, PARSE));
    printf("Preprocess time   %s\n", Print_clock(main_monitor_ptr, PREPROCESS));
    printf("Translate time    %s\n", Print_clock(main_monitor_ptr, TRANSLATE));
    printf("Build time        %s\n", Print_clock(main_monitor_ptr, BUILD));
    printf("Decision time     %s\n", Print_clock(main_monitor_ptr, DSAT));
    printf("--------------------GENERAL\n");
    printf("Consistency checks         %d\n", Get_param(main_monitor_ptr, REC_CALLS));
    printf("Failed consistency checks  %d\n", Get_param(main_monitor_ptr, MODFAIL_NUM));
    printf("Early Pruning (calls)      %d\n", Get_param(main_monitor_ptr, EP_CALLS));
    printf("Early Pruning (backtracks) %d\n", Get_param(main_monitor_ptr, EPFAIL_NUM));
    printf("--------------------DAVIS PUTNAM\n");
    printf("Unit propagations  %d\n", Get_param(main_monitor_ptr, UNIT_NUM));
    printf("Pure literals      %d\n", Get_param(main_monitor_ptr, PURE_NUM));
    printf("Splits             %d\n", Get_param(main_monitor_ptr, SPLIT_NUM));
    printf("  on modal atoms   %d\n", Get_param(main_monitor_ptr, SPLIT_MOD_NUM));
    printf("  on added var.    %d\n", Get_param(main_monitor_ptr, SPLIT_DEP_NUM));
    printf("Assignments found  %d\n", Get_param(main_monitor_ptr, ASSIGN_NUM));
    printf("Bactracks          %d\n", Get_param(main_monitor_ptr, PROPFAIL_NUM));
    printf("--------------------CNF CONVERSION\n");
    printf("Added variables   %d\n", Laa(dag_ptr) - Lma(dag_ptr));
    if (Use_memoS()) {
      float t_ratio;

      printf("--------------------CACHING\n");
      printf("Cache accesses      %d\n", Stats_memoS(MEMO_ACCESS));
      printf("Cache successes     %d\n", Stats_memoS(MEMO_SUCCESS));
      if (Stats_memoS(MEMO_STORAGE))
        t_ratio = 
          (float)(Stats_memoS(MEMO_TRASHING)) / (float)(Stats_memoS(MEMO_STORAGE));
      else
        t_ratio = 0.00;
      printf("Items stored        %d (trashing ratio %.2f)\n", 
	     Stats_memoS(MEMO_STORAGE), t_ratio);

      printf("Dependencies tried  %d\n", Stats_memoS(DEP_ACCESS));
      printf("Dependencies found  %d\n", 
	     (Stats_memoS(DEP_SUCCESS_PRUNE) + Stats_memoS(DEP_SUCCESS_UNSAT)));
      printf("   removed alphas   %d\n", Stats_memoS(DEP_SUCCESS_PRUNE));
      printf("   inconsistencies  %d\n", Stats_memoS(DEP_SUCCESS_UNSAT));
      if (Stats_memoS(DEP_STORAGE))
        t_ratio = 
          (float)(Stats_memoS(DEP_TRASHING)) / (float)(Stats_memoS(DEP_STORAGE));
      else
        t_ratio = 0.00;
      printf("Dependecies stored  %d (trashing ratio %.2f)\n", 
	     Stats_memoS(DEP_STORAGE), t_ratio);
    }
    printf("--------------------MEMORY\n");
    printf("Maximum memory allocated by SATO               %d Kb\n", 
	   (Get_param(main_monitor_ptr, MAX_SATO_MEMORY) /1024));
    if (Use_memoS()) {
      printf("Total memory allocated by CACHE + DEPENDENCIES %d Kb\n", (Stats_memoS(MEMORY) / 1024));
    }
    if (TIMEOUT == -1) printf("TIMEOUT WAS REACHED!!!!\n");
  } else {
    if (TIMEOUT != -1) {
      printf("%s %s %d %d %d %d %d %d %d %d %d %d %d %d ",
	     ((TIMEOUT == -1) ? "B" : (res ? "S" : "U")), 
	     Print_clock(main_monitor_ptr, DSAT), 
	     Get_param(main_monitor_ptr, REC_CALLS),
	     Get_param(main_monitor_ptr, MODFAIL_NUM),
	     Get_param(main_monitor_ptr, EP_CALLS),
	     Get_param(main_monitor_ptr, EPFAIL_NUM),
	     Get_param(main_monitor_ptr, UNIT_NUM),
	     Get_param(main_monitor_ptr, PURE_NUM),
	     Get_param(main_monitor_ptr, SPLIT_NUM),
	     Get_param(main_monitor_ptr, SPLIT_MOD_NUM),
	     Get_param(main_monitor_ptr, SPLIT_DEP_NUM),
	     Get_param(main_monitor_ptr, ASSIGN_NUM),
	     Get_param(main_monitor_ptr, PROPFAIL_NUM),
	     (Laa(dag_ptr) - Lma(dag_ptr)));
      if (Use_memoS()) {
        float t_ratio, t_ratio2;
        if (Stats_memoS(MEMO_STORAGE))
          t_ratio = 
            (float)(Stats_memoS(MEMO_TRASHING)) / (float)(Stats_memoS(MEMO_STORAGE));
        else
          t_ratio = 0.00;        
        if (Stats_memoS(DEP_STORAGE))
          t_ratio2 = 
            (float)(Stats_memoS(DEP_TRASHING)) / (float)(Stats_memoS(DEP_STORAGE));
        else
          t_ratio2 = 0.00;
	printf("%d %d %d %.2f %d %d %d %d %.2f ",
	       Stats_memoS(MEMO_ACCESS),
	       Stats_memoS(MEMO_SUCCESS),
	       Stats_memoS(MEMO_STORAGE),
	       t_ratio,
	       Stats_memoS(DEP_ACCESS),
	       Stats_memoS(DEP_SUCCESS_PRUNE),
	       Stats_memoS(DEP_SUCCESS_UNSAT),
	       Stats_memoS(DEP_STORAGE),
	       t_ratio2);
      }
      printf("%d ",
	     (Get_param(main_monitor_ptr, MAX_SATO_MEMORY) / 1024));
      if (Use_memoS()) {
	printf("%d ",
	       (Stats_memoS(MEMORY) / 1024));
      }
      printf("\n");
    }
  }

  return;
}

/* Setting defaults and parsing modal/preprocessing data only */
void Preparse_cmd_line(int argc, char **argv)
{
    int i, temp;

    CNF_CONV = STD_CONVERSION;  /* Default is: non-optimized conversion */
    NNF_CONV = 0;               /* Default is: no pushing of negations */
    LANGUAGE = NCDE;            /* Default is: restricted language (iff,or,and,not) */

    for(i = 1; i < argc; i++) 
      if (argv[i][0] == '-')  {
	temp = str_int(argv[i], 2);

	if (argv[i][1] == 'k') {
	  CNF_CONV = temp;
	} else if (argv[i][1] == 'n') {
	  NNF_CONV = 1;
	} else if (argv[i][1] == 'f') {
	  LANGUAGE = temp;
	} 
      }

    /* When the optimized conversion is requested... */
    if (CNF_CONV == OPT_CONVERSION) {
      NNF_CONV = 1;                 /* Negations MUST be pushed down */
      LANGUAGE = NCDE;              /* The language MUST allow for equivalences */
      PURE_HEUR = 1;                /* Pure literal is allowed on dependent variables */
    } else
      PURE_HEUR = 0;                /* Pure literal rule is disabled */

    return;
}

void Purge_comments()
{
  int ch;

  /* Purge lines of comments */
  while ((ch = getc(stdin)) == '%')
    ch = skip_chars_until(stdin, '\n');
  /* Rstore first non-comment character read */
  ch = ungetc(ch,stdin);
  
  return;
}
  
